""""
Usage:
pixie16-scope

Program to tune parameter in almost real time.
The program can show a "live" MCA spectra and current traces.

TODO:
* add showing history and option to go back to setting from a run in the history
  probably should create a setting class that gets stored with the history data and has a get() and apply() method that talks to the pixie16 directly
* show plots for FT, E, CFD (refactor code from binary browser, both program shoudl to the same)

"""

import cmd
from collections import OrderedDict
from pathlib import Path

import numpy as np
from matplotlib import pyplot as plt
from rich import print

import pixie16 as px


def validate(value, to_type, name):
    try:
        return to_type(value)
    except ValueError:
        print(f"[red]Error[/]Cannot convert {name} to {to_type}: {value}")
        print("Did not change value!")
        return


class MyCmd(cmd.Cmd):
    intro = (
        "Welcome to the pixie16 scope program\n\n"
        "Type 'help' to see the available commands.\n\n"
        "You need to set channel and especially mca_channel and load a settings file.\n"
    )

    prompt = "Pixie16>"
    CURSOR_KEYS = [
        "".join(chr(c) for c in [27, 91, 65]),  # cursor up
        "".join(chr(c) for c in [27, 91, 66]),  # cursor down
    ]

    def __init__(self) -> None:
        super().__init__()
        self.channel: int = 0
        self.module: int = 0  # should not change during runs
        self.MCA_data = np.zeros(shape=32768)
        self.traces = []
        self.old_data = []
        self.setting_file = None
        self.mca_runtime: float = 2
        self.trace_runtime: float = 2
        self.mca_channel: int = 0
        self.mca_normalize = False
        self.mca_rebin = 1

        self.fig, (self.ax_mca, self.ax_traces) = plt.subplots(nrows=1, ncols=2)

        # create some dummy data
        x = np.linspace(0, 10, 100)
        y = np.sin(x)
        self.ax_traces.plot(x, y)
        plt.show(block=False)

    def postloop(self):
        """Call the matplotlib event loop."""
        self.fig.canvas.draw_idle()
        self.fig.canvas.flush_events()

    def precmd(self, line):
        """Remove cursor up/down keys (readline not working in windows)."""
        for key in MyCmd.CURSOR_KEYS:
            line = line.replace(key, "")
        return line

    def take_traces(self):
        """Take traces and return a 2d numpy array."""
        px.control.enable_trace_settings(
            [(self.module, self.mca_channel)], disable_CFD=False
        )
        data = px.control.take_list_mode_data(self.trace_runtime)[self.module]
        # convert to evenst
        reader = px.read.list_mode.StreamReader()
        parser = px.read.list_mode.ListModeDataReader(reader)
        parser.reader.put(data)
        events = parser.pop_all()
        traces = [e.trace for e in events if e.channel == self.mca_channel]
        traces = [
            t for t in traces if len(t)
        ]  # sort out empty trace, which will just return a ()
        return np.asarray(traces)

    def take_MCA_spectrum(self):
        """Take MCA spectrum of the `mca channel`."""
        raw_settings = {
            "channels": [(self.module, self.channel)],
            "CaptureHistogram": True,
            "CaptureTrace": False,
        }
        px.control.change_setting_from_dict(raw_settings)
        data = px.control.take_MCA_spectra(
            [(self.module, self.mca_channel)], duration=self.mca_runtime
        )
        return data[0]

    def rebin(self, data):
        return data.reshape(-1, self.mca_rebin).sum(axis=1)

    def emptyline(self):
        """Do not repeat the last command (default)."""
        pass

    def exit(self):
        print("Good Bye!")
        return True

    def do_quit(self, arg):
        return self.exit()

    def do_bye(self, arg):
        return self.exit()

    def do_exit(self, arg):
        return self.exit()

    def clear(self):
        """Clear all plots and saved data.

        Move last data into old_data.
        """
        self.old_data.append((self.MCA_data, self.traces, self.channel))
        self.MCA_data = np.zeros((16, 32768))
        self.traces = []
        if self.fig:
            self.ax_mca.clear()
            self.ax_traces.clear()

    def reset(self):
        """Reload last setting file."""
        if self.setting_file is None:
            print(
                f"[red]ERROR[/] Cannot reload setting file, none set (use load <file>)."
            )
            return
        px.control.write_settings_from_file(self.setting_file)
        print(f"[yellow]INFO[/]Loaded file {self.setting_file}.")

    def do_channel(self, args):
        """Change the channel that gets affected by setting changes."""
        value = validate(args, int, "channel number")
        if value is not None:
            self.channel = value
            print(
                f"[yellow]INFO[/] Will change settings on {self.channel} from now on."
            )

    def do_mca_channel(self, args):
        """Change the channel we are looking at (traces and MCA)."""
        value = validate(args, int, "channel number")
        if value is not None:
            self.mca_channel = value
            self.reset()
            print(f"[yellow]INFO[/] Will read data from channel {self.mca_channel}.")

    def do_module(self, args):
        """Change the module that gets affected by setting changes (in slot numbers)."""
        value = validate(args, int, "module number")
        if value is not None:
            px.control.exit_system()
            px.control.init_and_boot(modules=[value])
            print(
                f"[yellow]INFO[/] Will change settings on module {value} from now on."
            )

    def do_normalize(self, args):
        """Toggle normilzation of MCA data."""
        self.mca_normalize = not self.mca_normalize
        if self.mca_normalize:
            print("[yellow]INFO[/] Will now normalize the MCA data.")
        else:
            print("[yellow]INFO[/] Will not normalize the MCA data.")

    # Fast Trigger variables
    def do_FTL(self, args):
        """Set L for the fast trigger filter [in us]."""
        value = validate(args, float, "fast trigger L")
        if value is not None:
            px.control.set_channel_parameter(
                "TRIGGER_RISETIME", value, self.module, self.channel
            )
            self.reset()

    def do_FTG(self, args):
        """Set G for the fast trigger filter [in us]."""
        value = validate(args, float, "fast trigger G")
        if value is not None:
            px.control.set_channel_parameter(
                "TRIGGER_FLATTOP", value, self.module, self.channel
            )
            self.reset()

    def do_FTT(self, args):
        """Set the threshold for the fast trigger filter [in ADC units]."""
        value = validate(args, int, "fast trigger threshold")
        if value is not None:
            px.control.set_channel_parameter(
                "TRIGGER_THRESHOLD", value, self.module, self.channel
            )
            self.reset()

    # Energy variables
    def do_EL(self, args):
        """Set L for the fast trigger filter [in us]."""
        value = validate(args, float, "energy L")
        if value is not None:
            px.control.set_channel_parameter(
                "ENERGY_RISETIME", value, self.module, self.channel
            )
            self.reset()

    def do_EG(self, args):
        """Set G for the fast trigger filter [in us]."""
        value = validate(args, float, "energy G")
        if value is not None:
            px.control.set_channel_parameter(
                "ENERGY_FLATTOP", value, self.module, self.channel
            )
            self.reset()

    def do_EPeak(self, args):
        """Set the location where the pulse height is at a maximum."""
        value = validate(args, int, "energy peaksample")
        if value is not None:
            raw_settings = {
                "channel": (self.module, self.channel),
                "PeakSample": [value],
            }
            px.control.change_raw_setting_from_dict(raw_settings)
            self.reset()

    # CFD variables
    def do_CFD(self, args):
        """Sets the CFD threshold for fast triggers."""
        value = validate(args, int, "CFD threshold")
        if value is not None:
            px.control.set_channel_parameter(
                "CFDThresh", value, self.module, self.channel
            )
            self.reset()

    def do_set_channel_variable(self, args):
        """Set and arbitraty setting on the pixie16."""
        name, value = args.split()
        px.control.set_channel_parameter(name, value, self.module, self.channel)

    def do_runtime(self, args):
        """Set the runtime for each MCA run."""
        runtime = validate(args, float, "MCA run time")
        if runtime is not None:
            self.mca_runtime = runtime
            print(
                f"[yellow]INFO[/] Using a runtime of {runtime} s for the MCA runs from now on."
            )

    def do_truntime(self, args):
        """Set the runtime for each MCA run."""
        truntime = validate(args, float, "Traces run time")
        if truntime is not None:
            self.trace_runtime = truntime
            print(
                f"[yellow]INFO[/] Using a runtime of {truntime} s for the traces from now on."
            )

    def do_traces(self, args):
        """Take trace data and display it.

        TODO:
        * option to exchange a bad one that got picked
        * add plots for FT, E, CFD
        * add showing in FPGA cycles instead of 2ns

        The above TOOD's could also be handeled by buttons in
        the QT window through matplotlib's widgets.
        """
        traces = self.take_traces()
        self.traces.extend(traces)

        if not self.traces:
            print(f"[red]ERROR[/] No traces for channel {self.channel}.")
            return

        L = len(traces[0])
        x = np.linspace(0, 2 * L, L)

        # calculate amplitudes
        amplitudes_and_traces = [(t.max() - t[0:50].mean(), t) for t in self.traces]
        amplitudes_and_traces.sort(key=lambda x: x[0])

        # pick 10 equally spaces in amplitudes
        N = np.linspace(0, 1, 10) * (len(self.traces) - 1)
        N = sorted({int(n) for n in N})
        traces = [amplitudes_and_traces[n][1] for n in N]

        self.ax_traces.clear()

        for t in traces:
            self.ax_traces.plot(x, t)

        self.ax_traces.set_xlabel("Time [ns]")
        self.ax_traces.set_ylabel("ADC units")
        self.fig.canvas.draw_idle()

    def do_mca(self, args):
        """Take MCA data and display it."""
        data = self.take_MCA_spectrum()
        self.MCA_data += data

        tmp = self.rebin(self.MCA_data)

        self.ax_mca.clear()
        if self.mca_normalize:
            tmp = tmp / tmp.sum()

        self.ax_mca.plot(tmp, drawstyle="steps-mid", color="blue")
        self.ax_mca.set_xlabel("Channels")
        self.ax_mca.set_ylabel("Counts/channel")
        self.fig.canvas.draw_idle()

    def do_scope(self, args):
        """View traces calculated from the channel trigger of 'mca_channel'."""
        # We need to change run settings to trigger on mca channel
        # We need to check which channels are valid for the pixie16
        # We need to plot traces only for channels with traces
        with px.control.temporary_settings():
            channels = [(self.module, i) for i in range(16) if i != self.mca_channel]
            ch_trig = OrderedDict(
                {
                    "channels": channels,
                    "TrigConfig0": 0x0000_0040,  # source signal is channel 9 and use FTIN_or
                    "TrigConfig1": 0,
                    "TrigConfig2": self.mca_channel << 28,
                    "TrigConfig3": 0,
                    "FastTrigSelect": "external",
                    "EnableChannelVal": False,
                }
            )

            px.control.change_setting_from_dict(ch_trig)
            custom_mask = OrderedDict(
                {
                    "channels": self.mca_channel,
                    "MultiplicityMaskL": 1 << self.mca_channel,
                    "MultiplicityMaskH": 0x0040_0000,
                }
            )
            px.control.change_setting_from_dict(custom_mask)
            px.control.enable_trace_settings(
                [(self.module, i) for i in range(16)], disable_CFD=False
            )
            data = px.control.take_list_mode_data(self.trace_runtime)[self.module]
            # convert to events
            reader = px.read.list_mode.StreamReader()
            parser = px.read.list_mode.ListModeDataReader(reader)
            parser.reader.put(data)
            events = parser.pop_all()
            tr = [e.trace for e in events]
            select = np.linspace(0, len(tr), 20)
            traces = tr[select]
            self.ax_traces.clear()

            for t in traces:
                self.ax_traces.plot(x, t)

            self.ax_traces.set_xlabel("Time [ns]")
            self.ax_traces.set_ylabel("ADC units")
            self.fig.canvas.draw_idle()

    def do_rebin(self, args):
        """Change the MCA rebin factor. This is given in 2^N with a default of 0."""
        value = validate(args, int, "rebin factor")

        if value is not None:
            self.mca_rebin = 2**value
            print(f"[yellow]INFO[/] rebinning MCA by 2**{value}.")

    def do_plots(self, args):
        """Reopen the plotting window in case it got closed."""
        self.fig.show()

    def do_load(self, args):
        """Load a settings file."""

        path = validate(args, Path, "path to load settings file")
        # call load setting file

        if path is None:
            print(f"[red]ERROR[/] cannot load file {path}")
            return

        px.control.write_settings_from_file(path)
        self.setting_file = path


def main():
    px.control.init_and_boot()
    MyCmd().cmdloop()
    px.control.exit_system()


if __name__ == "__main__":
    main()



import pytest

import pixie16


def test_setting_bits():
    number = 0
    start_bit = 16
    num_bits = 1
    value = 1
    result = pixie16.control._set_bits(number, start_bit, num_bits, value)
    assert result == 0b1_0000_0000_0000_0000

    number = 0
    start_bit = 2
    num_bits = 3
    value = 7
    result = pixie16.control._set_bits(number, start_bit, num_bits, value)
    assert result == 0b0_0000_0000_0001_1100


def test_masks():
    channels = [3, 4, 9]
    offset = 0

    result = pixie16.control._create_coincidence_masks(channels, offset)
    assert result == 0b0000_0010_0001_1000

    channels = [3, 4, 9]
    offset = 2

    result = pixie16.control._create_coincidence_masks(channels, offset)
    assert result == 0b0000_1000_0110_0000


def test_coincidence():
    settings = {
        "MultiplicitySelf": [4, 5, 9],
        "NrOfCoincidencesSelf": 2,
        "ChannelValidationSource": "coincidence",
    }

    result = pixie16.control._handle_coincidences(settings, 3)

    assert result["MultiplicityMaskL"][0] == 0b0000_0010_0011_0000
    assert len(result["MultiplicityMaskL"]) == 3
    assert result["MultiplicityMaskH"][0] == 0x0080_0000

    assert "MultiplicitySelf" not in settings
    assert "NrOfCoincidencesSelf" not in settings
    assert "ChannelValidationSource" not in settings


def test_expand_values():
    tmp = "a"

    result = pixie16.control._expand_values(tmp, 3)
    assert result == ["a", "a", "a"]

    tmp = 1.1

    result = pixie16.control._expand_values(tmp, 4)
    assert result == [1.1, 1.1, 1.1, 1.1]

    tmp = 42

    result = pixie16.control._expand_values(tmp, 5)
    assert result == [42, 42, 42, 42, 42]


def test_expand_list():
    tmp = []

    result = pixie16.control._expand_list(tmp, 3)
    assert result == [[], [], []]

    tmp = [1, 2, 3]

    result = pixie16.control._expand_list(tmp, 3)
    assert result == [[1, 2, 3], [1, 2, 3], [1, 2, 3]]



"""Read list mode data

The file provides a python based list mode data parser.

The input can either come from a data stream or from one or several
binary files (in case a large binary data set is split across multiple
files).

We keep all the data in numpy arrays, whic are either memmaped or
pre-allocated, to make data access fast.

"""

from collections import defaultdict
from collections.abc import Iterable
from itertools import islice
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd

try:
    import cbitstruct as bitstruct
except ModuleNotFoundError:
    import bitstruct

import msgspec

# from rich import print


class Event(msgspec.Struct):
    """Data storage for a single event that has attribute access.

    This is similar to a namedtuple, but the msgspec implemenation is faster.
    """

    channel: int = -1
    crate: int = -1
    slot: int = -1
    timestamp: float = -1.0
    CFD_fraction: int = -1
    energy: int = -1
    trace: int = -1
    CFD_error: int = -1
    pileup: int = -1
    trace_flag: int = -1
    Esum_trailing: int = -1
    Esum_leading: int = -1
    Esum_gap: int = -1
    baseline: float = -1.0
    QDCSum0: int = -1
    QDCSum1: int = -1
    QDCSum2: int = -1
    QDCSum3: int = -1
    QDCSum4: int = -1
    QDCSum5: int = -1
    QDCSum6: int = -1
    QDCSum7: int = -1
    ext_timestamp: float = -1.0
    chunk_timestamp: float = -1  # not a pixie16 field

    def to_list(self):
        return (getattr(self, f) for f in self.__struct_fields__)


# The bit structure to unpack the pixie binary format
EVENTS_PARTS = {
    "header": (
        "b1u14u5u4u4u4u32u3u13u16b1u15u16",
        (
            "pileup",
            "Event Length",
            "Header Length",
            "crate",
            "slot",
            "channel",
            "EVTTIME_LO",
            "CFD trigger source bits",
            "CFD Fractional Time",
            "EVTTIME_HI",
            "trace_flag",
            "Trace Length",
            "energy",
        ),
    ),
    "energy": (
        "u32" * 3 + "f32",
        (
            "Esum_trailing",
            "Esum_leading",
            "Esum_gap",
            "baseline",
        ),
    ),
    "Qsums": (
        "u32" * 8,
        (
            "QDCSum0",
            "QDCSum1",
            "QDCSum2",
            "QDCSum3",
            "QDCSum4",
            "QDCSum5",
            "QDCSum6",
            "QDCSum7",
        ),
    ),
    "Ext Time": (
        "u32p16u16",
        (
            "Ext_TS_Lo",
            "Ext_TS_Hi",
        ),
    ),
}

# We now build all possible combinations out of these parts that can
# show up in list mode data We also add a "0" option to just read the
# first row, since this row contains the header length we need to read
HEADER_FORMATS = bitstruct.CompiledFormatDict(
    EVENTS_PARTS["header"][0], EVENTS_PARTS["header"][1]
)

# when more information is in the header, we look it up using the
# following data structure. The key into this dictionary is how many
# 32-bit words we still need to read
EXTRA_HEADER_FORMATS = {
    2: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["Ext Time"][0],
        EVENTS_PARTS["Ext Time"][1],
    ),
    4: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["energy"][0],
        EVENTS_PARTS["energy"][1],
    ),
    8: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["Qsums"][0],
        EVENTS_PARTS["Qsums"][1],
    ),
    12: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["energy"][0] + EVENTS_PARTS["Qsums"][0],
        EVENTS_PARTS["energy"][1] + EVENTS_PARTS["Qsums"][1],
    ),
    6: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["energy"][0] + EVENTS_PARTS["Ext Time"][0],
        EVENTS_PARTS["energy"][1] + EVENTS_PARTS["Ext Time"][1],
    ),
    10: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["Qsums"][0] + EVENTS_PARTS["Ext Time"][0],
        EVENTS_PARTS["Qsums"][1] + EVENTS_PARTS["Ext Time"][1],
    ),
    14: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["energy"][0]
        + EVENTS_PARTS["Qsums"][0]
        + EVENTS_PARTS["Ext Time"][0],
        EVENTS_PARTS["energy"][1]
        + EVENTS_PARTS["Qsums"][1]
        + EVENTS_PARTS["Ext Time"][1],
    ),
}


class EmptyError(Exception):
    """We reached the end of the last file."""


class LeftoverBytesError(Exception):
    """To be raised if a partial event is left in the byte stream/file."""


class FileReader:
    """Read binary data from multiple files.

    Use mmap for fast data access to binary data on disc. The data can
    be distributed across several files.

    This class should  be used as a context manager:
    >>> files = ("file1.bin", "file2.bin")
    >>> with FileReader(files) as r:
    >>>     r.read(100)

    or one needs to call self.open_files() and self.close_files()
    manually.

    """

    def __init__(self, files: Iterable[Path]):
        self.current_file = 0  # which file do we need to read from
        self.files = files  # list of all files
        self.file_size = []
        self.position = 0  # relative to the current file
        self.mmap = []
        self.done = False  # have we reached the end of the last file?

    def open_files(self):
        for f in self.files:
            self.mmap.append(np.memmap(f, dtype=np.uint32, mode="r"))
            self.file_size.append(len(self.mmap[-1]))

    def close_files(self):
        self.file_size = []
        self.file_objects = []
        self.mmap = []
        self.current_file = 0

    def __enter__(self):
        """Opens all files, check their sizes and open a mmap read_only connection."""
        self.open_files()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Closes all file descriptors and mmaps at the end of the context manager."""
        self.close_files()

    def read(self, size: int) -> bytes:
        """Retrieve the next `size` bytes from the files and advance the file position.

        Parameters
        ----------
        size
            How many bytes should be returned.

        Raises
        ------
        EmptyError
            When data of the requested size is not available in the data stream.

        """

        if self.done:
            raise EmptyError()

        end = size + self.position
        if end <= self.file_size[self.current_file]:
            out = self.mmap[self.current_file][self.position : end]
            self.position += size
            return out

        tmp = self.mmap[self.current_file][self.position :]
        to_read = size - len(tmp)

        self.current_file += 1
        self.position = 0

        if self.current_file == len(self.files):
            self.done = True
            if len(tmp) == 0:
                raise EmptyError
            raise LeftoverBytesError

        rest = self.read(to_read)
        if tmp.size > 0:
            return np.concatenate((tmp, rest))
        else:
            return rest


class StreamReader:
    """Handle streamed binary data.

    Provides the same interface as FileReader and can be used as a direct replacement.

    New data can be added using the `put` method.

    Internally we store the data in a np.array of np.uint32.

    We make the array large enough to hold the data and have the
    buffer grow and shrink as needed.

    This is similar to a ring buffer, but we just allocate twice the buffer
    size memory and then keep track of the left and right position in
    the buffer and shift the data to the far left every now and then.

    """

    def __init__(self, initial_buffer_size=25_000_000) -> None:
        self.buffer_size = initial_buffer_size
        self.buffer_size_orig = self.buffer_size
        self.buffer = np.empty(2 * self.buffer_size, dtype=np.uint32)
        self.left = 0
        self.right = 0

    def put(self, input: np.ndarray) -> None:
        """Add data to the buffer.

        Also increase/decrease the buffer size as needed.
        """

        L = input.size
        current = self.right - self.left

        buffer_size_change = False
        # adjust buffer so that it will be large enough for the new data
        while current + L >= self.buffer_size:
            self.buffer_size *= 2
            buffer_size_change = True

        # Can we shrink the buffer?
        while (self.buffer_size > self.buffer_size_orig) and (
            current + L < self.buffer_size // 2
        ):
            self.buffer_size = self.buffer_size // 2
            buffer_size_change = True

        # check if we need to shift the buffer
        if self.left > self.buffer_size or buffer_size_change:
            self.adjust_and_shift_buffer()

        # add input
        self.buffer[self.right : self.right + L] = input
        self.right += L

    def adjust_and_shift_buffer(self) -> None:
        """Create a new buffer and copy the old buffer to beginning.

        This also takes care of changing the size of the buffer
        """
        if (self.left == 0) and (self.buffer.size == self.buffer_size):
            return

        current = self.right - self.left
        tmp = np.empty(2 * self.buffer_size, dtype=np.uint32)
        tmp[:current] = self.buffer[self.left : self.right]
        self.buffer = tmp
        self.left = 0
        self.right = current

    def read(self, size: int) -> np.ndarray:
        """Return `size` 32-bit words from the buffer as a numpy array."""
        current = self.right - self.left
        if size <= current:
            out = self.buffer[self.left : self.left + size]
            self.left += size
            return out
        raise EmptyError

    def __len__(self):
        """Return the amount of data in the buffer."""
        return self.right - self.left


class ListModeDataReader:
    def __init__(self, reader: Union[FileReader, StreamReader]):
        self.reader = reader

    def pop(self) -> Event:
        """Pop the top event from queue"""

        # try to get all the words for the next event
        # if the complete next event isn't in queue an exception is thrown but no data is discarded
        event_header = HEADER_FORMATS.unpack(self.reader.read(4).byteswap().tobytes())
        header_length = event_header["Header Length"]
        extra_header_length = header_length - 4
        event_length = event_header["Event Length"]

        # it's now save to read both header and trace
        if extra_header_length > 0:
            header_words = self.reader.read(extra_header_length).byteswap().tobytes()
            # unpack words/trace into their field values
            try:
                event_values = EXTRA_HEADER_FORMATS[extra_header_length].unpack(
                    header_words
                )
                event_header.update(**event_values)
            except KeyError:
                print("Something went wrong with the binary data.")
                raise

        trace_length = event_length - header_length
        if trace_length:
            trace_data = self.reader.read(trace_length).tobytes()
            trace = np.frombuffer(trace_data, f"({trace_length},2)<u2", count=1)[
                0, :, :
            ]
            trace = trace.flatten()
        else:
            trace_data = None
            trace = ()

        # calculate event values that need to be derived from field values
        CFD_bits = event_header.pop("CFD trigger source bits")
        CFD_error = CFD_bits == 7
        CFD_fractional_time = event_header.pop("CFD Fractional Time")
        if CFD_bits == 7:
            CFD_fraction = 0
        else:
            CFD_fraction = ((CFD_bits - 1) + CFD_fractional_time / 8192) * 2

        timestamp = (
            event_header.pop("EVTTIME_LO") + event_header.pop("EVTTIME_HI") * 2**32
        ) * 10

        if "Ext_TS_Lo" in event_header:
            ext_timestamp = (
                event_header.pop("Ext_TS_Lo") + event_header.pop("Ext_TS_Hi") * 2**32
            )

            event_header.update({"ext_timestamp": ext_timestamp})

        for key in ["Event Length", "Header Length", "Trace Length"]:
            event_header.pop(key, None)

        event_header.update(
            {
                "timestamp": timestamp,
                "CFD_fraction": CFD_fraction,
                "trace": trace,
                "CFD_error": CFD_error,
            }
        )

        return Event(**event_header)

    def iterevents(self):
        """Will raise an exception to exit loop."""
        while True:
            yield self.pop()

    def pop_all(self):
        out = []
        while True:
            try:
                event = self.pop()
            except (EmptyError, LeftoverBytesError):
                break
            out.append(event)
        return out


def events_from_files_generator(files):
    """Yields events from a list of binary files"""
    try:
        with FileReader(files) as r:
            reader = ListModeDataReader(r)
            while not reader.reader.done:
                yield from reader.iterevents()
    except (EmptyError, LeftoverBytesError):
        return


def to_tuple(obj):
    if isinstance(obj, (str, Path)):
        obj = [obj]
    else:
        try:
            obj = tuple(obj)
        except TypeError:
            obj = obj
    return obj


def read_list_mode_data(files, buffer_size=int(1e9)):
    """Loads a list of binary files into a pandas DataFrame"""
    files = to_tuple(files)
    events = (e.to_list() for e in events_from_files_generator(files))
    return pd.DataFrame.from_records(events, columns=Event.__struct_fields__)


def read_list_mode_data_as_events(files, buffer_size=int(1e9), max_size=None):
    """Returns a list of at most max_size events from files"""
    files = to_tuple(files)
    gen = events_from_files_generator(files)
    return list(islice(gen, max_size))


def sort_events_by_channel(events, channel_list=None):
    """Sort events by channel.

    Takes the output of `read_list_mode_data_as_events` and sorts them
    into a dictionary that has the channel number as key and as value
    a list of events (still including the channel number).

    Parameters
    ----------
    events
        List of events as returned by `read_list_mode_data_as_events`
    channel_list
        Optional list of channels to include. This can help to reduce
        the amount of data.

    """

    out = defaultdict(list)

    for e in events:
        ch = e.channel
        if channel_list is not None and ch in channel_list:
            out[ch].append(e)
        else:
            out[ch].append(e)

    return out
